###################################################################################
#
#    Copyright (C) 2020 Cetmix OÜ
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU LESSER GENERAL PUBLIC LICENSE as
#    published by the Free Software Foundation, either version 3 of the
#    License, or (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU LESSER GENERAL PUBLIC LICENSE for more details.
#
#    You should have received a copy of the GNU LESSER GENERAL PUBLIC LICENSE
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
###################################################################################

import base64

from odoo import _, models, tools
from odoo.exceptions import UserError
from odoo.tools import pycompat


#################
# Mail Template #
#################
class MailTemplate(models.Model):
    _inherit = "mail.template"

    # -- Cetmix. If message is quoted put signature before quoted text
    def generate_email(self, res_ids, fields=None):  # noqa
        """Generates an email from the template for given the given model based on
        records given by res_ids.

        :param template_id: id of the template to render.
        :param res_id: id of the record to use for rendering the template (model
                       is taken from template definition)
        :returns: a dict containing all relevant fields for creating a new
                  mail.mail entry, with one extra key ``attachments``, in the
                  format [(report_name, data)] where data is base64 encoded.
        """
        # Cetmix. Check context.
        if not self._context.get("default_wizard_mode", False) in ["quote", "forward"]:
            return super(MailTemplate,
                         self).generate_email(res_ids=res_ids, fields=fields)

        # Get signature location
        signature_location = self._context.get("signature_location", False)

        # No signature
        if signature_location == "n":
            return super(MailTemplate, self.with_context(
                signature=False
            )).generate_email(res_ids=res_ids, fields=fields)
        # After quote
        if signature_location == "a":
            return super(MailTemplate,
                         self).generate_email(res_ids=res_ids, fields=fields)

        self.ensure_one()
        multi_mode = True
        if isinstance(res_ids, pycompat.integer_types):
            res_ids = [res_ids]
            multi_mode = False
        if fields is None:
            fields = [
                "subject",
                "body_html",
                "email_from",
                "email_to",
                "partner_to",
                "email_cc",
                "reply_to",
                "scheduled_date",
            ]

        res_ids_to_templates = self.get_email_template(res_ids)

        # templates: res_id -> template; template -> res_ids
        templates_to_res_ids = {}
        for res_id, template in res_ids_to_templates.items():
            templates_to_res_ids.setdefault(template, []).append(res_id)

        results = dict()

        # Cetmix. Save signature
        signature = self._context.get("signature", False)
        for template, template_res_ids in templates_to_res_ids.items():
            Template = self.env["mail.template"]
            # generate fields value for all res_ids linked to the current template
            if template.lang:
                Template = Template.with_context(lang=template._context.get("lang"))

            for field in fields:
                # Cetmix. Set signature in context to False
                # to avoid rendering in template
                Template = Template.with_context(
                    safe=field in {"subject"}, signature=False
                )
                generated_field_values = Template.render_template(
                    getattr(template, field),
                    template.model,
                    template_res_ids,
                    post_process=(field == "body_html"),
                )
                for res_id, field_value in generated_field_values.items():
                    # Cetmix. Insert signature before <blockquote...
                    new_body = False
                    if field == "body_html":
                        quote_index = field_value.find("<blockquote")
                        if quote_index:
                            new_body = "{}{}{}".format(
                                field_value[:quote_index],
                                signature,
                                field_value[quote_index:],
                            )  # legacy mode
                    results.setdefault(
                        res_id, dict()
                    )[field] = (new_body if new_body else field_value)
            # compute recipients
            if any(field in fields for field in ["email_to", "partner_to", "email_cc"]):
                results = template.generate_recipients(results, template_res_ids)
            # update values for all res_ids
            for res_id in template_res_ids:
                values = results[res_id]
                # body: add user signature, sanitize
                if "body_html" in fields and template.user_signature:
                    signature = self.env.user.signature
                    if signature:
                        values["body_html"] = tools.append_content_to_html(
                            values["body_html"], signature, plaintext=False
                        )
                if values.get("body_html"):
                    values["body"] = tools.html_sanitize(values["body_html"])
                # technical settings
                values.update(
                    mail_server_id=template.mail_server_id.id or False,
                    auto_delete=template.auto_delete,
                    model=template.model,
                    res_id=res_id or False,
                    attachment_ids=[attach.id for attach in template.attachment_ids],
                )

            # Add report in attachments: generate once for all template_res_ids
            if template.report_template:
                for res_id in template_res_ids:
                    attachments = []
                    report_name = self.render_template(
                        template.report_name, template.model, res_id
                    )
                    report = template.report_template
                    report_service = report.report_name

                    if report.report_type not in ["qweb-html", "qweb-pdf"]:
                        raise UserError(
                            _("Unsupported report type %s found.") % report.report_type
                        )
                    result, format_ = report.render_qweb_pdf([res_id])  # noqa

                    # TODO in trunk, change return format to
                    #  binary to match message_post expected format
                    result = base64.b64encode(result)
                    if not report_name:
                        report_name = "report." + report_service
                    ext = "." + format_
                    if not report_name.endswith(ext):
                        report_name += ext
                    attachments.append((report_name, result))
                    results[res_id]["attachments"] = attachments

        return multi_mode and results or results[res_ids[0]]
